/*
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*
 * $Log: httpd.c,v $
 * Revision 1.10  2002/06/26 17:29:49  harald
 * First pre-release with 2.4 stack
 *
 */

#include <string.h>

#include <sys/device.h>

#include <sys/timer.h>
#include <sys/heap.h>

#include <arpa/inet.h>
#include <net/if_var.h>
#include <netinet/sostream.h>

#include <sys/print.h>

#include <fs/uromfs.h>
#include "dencode.h"

#include <pro/httpd.h>

/*!
 * \addtogroup xgHTTPD
 */
/*@{*/

static struct {
    char* ext;
    char* type;
} mimeTypes[] = {
    { ".txt",  "text/plain; charset=iso-8859-1" },
    { ".html", "text/html; charset=iso-8859-1" },
    { ".gif",  "image/gif" },
    { ".jpg",  "image/jpeg" }
};


/*!
 * \brief Send top lines of a standard HTML header.
 *
 * Sends HTTP and Server version lines.
 *
 * \param sostream Socket stream of the connection, previously
 *                 created by calling NutSoStreamCreate().
 * \param req      The associated client request.
 * \param status   Response status, error code or 200, if no
 *                 error occured.
 * \param title    Error text, or OK, if no error occured.
 */
void NutHttpSendHeaderTop(NUTDEVICE *sostream, REQUEST *req, int status, char* title)
{
    NutPrintFormat(sostream, "HTTP/%d.%d %d %s\r\n",
                req->req_version / 10, req->req_version % 10, status, title);
    NutPrintString(sostream, "Server: Ethernut 2.0\r\n");
}

/*!
 * \brief Send bottom lines of a standard HTML header.
 *
 * Sends Content-Type, Content-Lenght and Connection lines.
 *
 * \param sostream  Socket stream of the connection, previously
 *                  created by calling NutSoStreamCreate().
 * \param mime_type Points to a string that specifies the content
 *                  type. Examples are "text/html", "image/png", 
 *                  "image/gif", "video/mpeg" or "text/css".
 *                  A null pointer is ignored.
 * \param bytes     Content length of the data following this
 *                  header. Ignored, if negative.
 */
void NutHttpSendHeaderBot(NUTDEVICE *sostream, char* mime_type, int bytes)
{
    if(mime_type)
        NutPrintFormat(sostream, "Content-Type: %s\r\n", mime_type);
    if(bytes >= 0)
        NutPrintFormat(sostream, "Content-Length: %d\r\n", bytes);
    NutPrintFormat(sostream, "Connection: close\r\n\r\n");
}

static char *err_file =
    "<HTML>"
    "<HEAD>"
    "<TITLE>%d %s</TITLE>"
    "</HEAD>"
    "<BODY>"
    "%d %s"
    "</BODY>"
    "</HTML>";

/*!
 * \brief Send a HTTP error response.
 *
 * A canned error file is used.
 *
 * \param sostream Socket stream of the connection, previously
 *                 created by calling NutSoStreamCreate().
 * \param req      Contains the HTTP request.
 * \param status   Error code to be returned.
 */
void NutHttpSendError(NUTDEVICE *sostream, REQUEST *req, int status)
{
    char *title;

    switch(status) {
    case 400:
        title = "Bad Request";
        break;
    case 401:
        title = "Unauthorized";
        break;
    case 404:
        title = "Not Found";
        break;
    case 500:
        title = "Internal Error";
        break;
    case 501:
        title = "Not Implemented";
        break;
    default:
        title = "Error";
        break;
    }

    NutHttpSendHeaderTop(sostream, req, status, title);
    if(status == 401) {
        char *cp = 0;
        char *realm = req->req_url;

        if((cp = strrchr(realm, '/')) != 0)
            *cp = 0;
        else
            realm = ".";
        NutPrintFormat(sostream, "WWW-Authenticate: Basic realm=\"%s\"", realm);
        if(cp)
            *cp = '/';
    }
    NutHttpSendHeaderBot(sostream, "text/html", -1);
    NutPrintFormat(sostream, err_file, status, title, status, title);

}

/*!
 * \brief Return the mime type description of a specified file name.
 *
 * The mime type returned is based on the file extension.
 *
 * \todo Function to register additional mime types. Currently only
 *       .txt, .html, .gif and .jpg are supported.
 *
 * \param name Name of the file.
 *
 * \return A pointer to a static string, containing the
 *         associated mime type description. If the extension
 *         is not registered, "text/plain; charset=iso-8859-1"
 *         is returned. If the filename is empty, then
 *         "text/hatml; charset=iso-8859-1" is returned.
 */
char *NutGetMimeType(char* name)
{
    int i;
    int fl;

    if(name == 0 || (fl = strlen(name)) == 0)
        return mimeTypes[1].type;    
    for(i = 0; i < sizeof(mimeTypes) / sizeof(*mimeTypes); i++)
        if(strcasecmp(&(name[fl - strlen(mimeTypes[i].ext)]), mimeTypes[i].ext) == 0)
            return mimeTypes[i].type;
    return mimeTypes[0].type;
}

/*!
 *
 */
static void NutHttpProcessFileRequest(NUTDEVICE *sostream, REQUEST *req)
{
    ROMFILE *fd;
    int n;
    char *data;
    int size;
    int written;

    /*
     * Validate authorization.
     */
    if(NutHttpAuthValidate(req)) {
        NutHttpSendError(sostream, req, 401);
        return;
    }

    /*
     * Process CGI.
     */
    if(strncasecmp(req->req_url, "cgi-bin/", 8) == 0) {
        NutCgiProcessRequest(sostream, req);
        return;
    }

    /*
     * Process file.
     */
    if((fd = NutRomFileOpen(req->req_url)) == 0) {
        u_char *index;
        u_short urll;

        urll = strlen(req->req_url);
        if((index = NutHeapAllocClear(urll + 12)) == 0) {
            NutHttpSendError(sostream, req, 500);
            return;
        }
        if(urll)
            strcpy(index, req->req_url);
        if(urll && index[urll - 1] != '/')
            strcat(index, "/");
        strcat(index, "index.html");
        if((fd = NutRomFileOpen(index)) == 0) {
            NutHeapFree(index);
            NutHttpSendError(sostream, req, 404);
            return;
        }
        NutHeapFree(index);
    }
    size = NutRomFileSize(fd);
    NutHttpSendHeaderTop(sostream, req, 200, "Ok");
    NutHttpSendHeaderBot(sostream, NutGetMimeType(req->req_url), size);
    if(req->req_method != METHOD_HEAD) {
        if(size > 512)
            size = 512;
        if((data = NutHeapAlloc(size)) != 0) {
            while((n = NutRomFileRead(fd, data, size)) != 0) {
                written = NutPrintBinary(sostream, data, n);
            }
            NutHeapFree(data);
            NutPrintFlush(sostream);
        }
    }
    NutRomFileClose(fd);
}

/*!
 *
 */
static char *NextWord(char *str)
{
    while(*str && *str != ' ' && *str != '\t')
        str++;
    if(*str)
        *str++ = 0;
    while(*str == ' ' || *str == '\t')
        str++;
    return str;
}

/*!
 *
 */
static REQUEST *CreateRequestInfo(void)
{
    return NutHeapAllocClear(sizeof(REQUEST));
}

/*!
 *
 */
static void DestroyRequestInfo(REQUEST *req)
{
    if(req->req_url)
        NutHeapFree(req->req_url);
    if(req->req_query)
        NutHeapFree(req->req_query);
    if(req->req_type)
        NutHeapFree(req->req_type);
    if(req->req_cookie)
        NutHeapFree(req->req_cookie);
    if(req->req_auth)
        NutHeapFree(req->req_auth);
    if(req->req_agent)
        NutHeapFree(req->req_agent);
    NutHeapFree(req);
}

/*!
 * \brief Process the next HTTP request.
 *
 * Waits for the next HTTP request on an established connection
 * and processes it.
 *
 * \param sostream Socket stream of the connection, previously
 *                 created by calling NutSoStreamCreate().
 */
void NutHttpProcessRequest(NUTDEVICE *sostream)
{
    REQUEST *req;
    char *method;
    char *path;
    char *line;
    char *protocol;
    char *cp;

    /*
     * Parse the first line of the request.
     * <method> <path> <protocol>
     */
    if((req = CreateRequestInfo()) == 0)
        return;
    if((method = NutHeapAlloc(256)) == 0) {
        DestroyRequestInfo(req);
        return;
    }
    NutDeviceGetLine(sostream, method, 256);

    /*
     * Parse remaining request headers.
     */
    if((line = NutHeapAlloc(256)) == 0) {
        NutHeapFree(method);
        DestroyRequestInfo(req);
        return;
    }
    for(;;) {
        if(NutDeviceGetLine(sostream, line, 256) <= 0)
            break;
        if(*line == 0)
            break;
        if(strncasecmp(line, "Authorization:", 14) == 0) {
            if(req->req_auth == 0) {
                cp = &line[14];
                while(*cp == ' ' || *cp == '\t')
                    cp++;
                if((req->req_auth = NutHeapAlloc(strlen(cp) + 1)) == 0)
                    break;
                strcpy(req->req_auth, cp);
            }
        }
        else if(strncasecmp(line, "Content-Length:", 15) == 0)
            req->req_length = 0 /*atoi(&line[15])*/;
        else if(strncasecmp(line, "Content-Type:", 13) == 0) {
            if(req->req_type == 0) {
                cp = &line[13];
                while(*cp == ' ' || *cp == '\t')
                    cp++;
                if((req->req_type = NutHeapAlloc(strlen(cp) + 1)) == 0)
                    break;
                strcpy(req->req_type, cp);
            }
        }
        else if(strncasecmp(line, "Cookie:", 7) == 0) {
            if(req->req_cookie == 0) {
                cp = &line[7];
                while(*cp == ' ' || *cp == '\t')
                    cp++;
                if((req->req_cookie = NutHeapAlloc(strlen(cp) + 1)) == 0)
                    break;
                strcpy(req->req_cookie, cp);
            }
        }
        else if(strncasecmp(line, "User-Agent:", 11) == 0) {
            if(req->req_agent == 0) {
                cp = &line[11];
                while(*cp == ' ' || *cp == '\t')
                    cp++;
                if((req->req_agent = NutHeapAlloc(strlen(cp) + 1)) == 0)
                    break;
                strcpy(req->req_agent, cp);
            }
        }
    }
    NutHeapFree(line);

    path = NextWord(method);
    protocol = NextWord(path);
    NextWord(protocol);

    if(strcasecmp(method, "GET") == 0)
        req->req_method = METHOD_GET;
    else if(strcasecmp(method, "HEAD") == 0)
        req->req_method = METHOD_HEAD;
    else if(strcasecmp(method, "POST") == 0)
        req->req_method = METHOD_POST;
    else {
        NutHeapFree(method);
        DestroyRequestInfo(req);
        NutHttpSendError(sostream, req, 501);
        return;
    }
    if(*path == 0 || *protocol == 0) {
        NutHeapFree(method);
        DestroyRequestInfo(req);
        NutHttpSendError(sostream, req, 400);
        return;
    }

    req->req_version = 10;
    if(strcasecmp(protocol, "HTTP/0.9") == 0)
        req->req_version = 9;

    if((cp = strchr(path, '?')) != 0) {
        *cp++ = 0;
        if((req->req_query = NutHeapAlloc(strlen(cp) + 1)) == 0) {
            NutHeapFree(method);
            DestroyRequestInfo(req);
            return;
        }
        strcpy(req->req_query, cp);
    }
    if((req->req_url = NutHeapAlloc(strlen(path) + 1)) == 0) {
        NutHeapFree(method);
        DestroyRequestInfo(req);
        return;
    }
    strcpy(req->req_url, path);

    NutHeapFree(method);

    if(NutDecodePath(req->req_url) == 0)
        NutHttpSendError(sostream, req, 400);
    else
        NutHttpProcessFileRequest(sostream, req);

    DestroyRequestInfo(req);
}

/*@}*/
